;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cello; -*-
;;;
;;; Copyright  2004 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :cello)

(defconstant *referenceDPI* 1440)

(let (
      (logical-dpi 96) ;;1440) 
      ; This is cello's internal dots per inch.  This value is germane only if size references are unqualified by a function call.
      ; Size references should always be qualified, as in (:pts 6), except when specifying pen widths.
      ; (Pen widths may pose a special case -- we may need to match screen pens to print pens.)
      
      (scan-resolution 300)                  
      ; This is the desired scan resolution, and the assumed resolution of all scans.
      ; Hypothetically, a scanner not capable of scanning at 300 dpi could make a big hash of this scheme.
      ; Rather than even pretend to support multiple resolutions within a study, for now we'll enforce 300 across the board.
      ; Dependencies on this spec can be identified by searching on scan-resolution.
      
      (logical-screen-resolution 96)         
      ; This is the internal logical screen resolution, which does _not_ have to equal the current LOGPIXELSX (LOGPIXELSY) value
      ; reported by GetDeviceCaps.  The original thought was that we could use this to rescale _all_ drawing on the fly.  Now that
      ; idea is being superseded by targetRes, but this functions (1) as a tacit targetRes for the outer window and (2) as a magic
      ; number to complicate debugging [we need to root out a few references in .bmp drawing, I think].
      
      ;;(printer-resolution 600)      ; /// improve #'cs-printer-resolution to bypass this.
      
      ;;(emf-resolution 600)
      
      )
  
  (declare (ignorable logical-dpi scan-resolution logical-screen-resolution printer-resolution))
  
  ; Notice the somewhat nonstandard naming convention:
  ; #'uInches takes logical inches and returns logical units (DPI)
  ; so, for instance, if logical-dpi = 1440, then (uInches 0.5) = 720.
  (defun uRound (number &optional (divisor 1))
    (multiple-value-bind (quotient remainder)
        (round number divisor)
      (declare (ignorable remainder))
      ;(assert (zerop remainder))
      ;(assert (zerop (mod quotient 15))) ;96ths
      quotient))
  

  (defun uDots (dots DPI)
    (uround (* dots logical-dpi) DPI))   ;only the first value will be used.
  
  (defun uInches (logical-inches)
    (uround (* logical-inches logical-dpi)))   ;only the first value will be used.
  
  (defun uIn (logical-inches)
    (uInches logical-inches))
  
  (defun uPoints (logical-points)
    (uDots logical-points  72))
  
  (defun uPts (logical-points)
    (uPoints logical-points))
  
  (defun u96ths (logical-96ths)
    (uDots logical-96ths 96))
  
  (defun u8ths (logical8ths)
    (uDots logical8ths 8))
  
  (defun u16ths (logical16ths)
    (uDots logical16ths 16))
  
  (defun u32nds (logical32nds)
    (uDots logical32nds 32))
  
  (defun u120ths (logical-120ths)
    (uDots logical-120ths 120))
  
  (defun cs-logical-dpi ()
    logical-dpi)
  
  (defsetf cs-logical-dpi cs-logical-dpi-setf)
  
  (defun cs-logical-dpi-setf (newValue)
    (setf logical-dpi newValue))
  
  (defun cs-scan-resolution ()
    scan-resolution)
  
  (defun cs-logical-screen-resolution ()
    logical-screen-resolution)
  
  )




(defmethod uCvt ((nn number) (units (eql :96ths)) )
  (u96ths nn))

(defmethod uCvt ((nn number) (units (eql :8ths)) )
  (u8ths nn))

(defmethod uCvt ((nn number) (units (eql :16ths)) )
  (u16ths nn))

(defmethod uCvt ((nn number) (units (eql :32nds)) )
  (u32nds nn))

(defmethod uCvt ((nn number) (units (eql :inches)) )
  (uInches nn))

(defmethod uCvt ((nn number) (units (eql :points)) )
  (uPoints nn))

(defmethod uCvt (other units)
  (declare (ignore units))
  other)

(defmethod uCvt ((nns cons) units)
  (cons (uCvt (car nns) units)
        (uCvt (cdr nns) units)))

(defmacro uCvt! (nn units)
  `(uCvt ,nn ,units))

(defun uv2 (x y uKey) (apply #'mkv2 (uCvt (list x y) uKey)))

;-----------------

(defun os-logical-screen-dpi ()
  (break "need (win:GetDeviceCaps (device-context (screen *cg-system*)) win:LOGPIXELSX))"))
   
#+no(defun browser-target-resolution ()
  (targetResolution (find-window :clinisys)))

; set to 96 because the code is trying to do rect-frames for the header before the window is init'ed.

(let ((current-target-resolution 96))  ;initialize when main window is created  
   
   (defun set-current-target-resolution (resolution)
     #+shh(trc "setting current-target-resolution to" resolution)
     (setf current-target-resolution resolution))
   
   (defun cs-current-target-resolution ()
     current-target-resolution)
   
   (defun cs-target-res ()
     current-target-resolution)
   
   (defmacro with-target-resolution ((new-resolution) &rest body)
     (let ((old-resolution (gensym))
           )
        `(let ((,old-resolution (cs-current-target-resolution))
               )
            (prog2
              (set-current-target-resolution ,new-resolution)
                (progn ,@body)
              (set-current-target-resolution ,old-resolution)
            ))))
   )


;converts screen pixels to logical pixels given the current target resolution OR OPTIONAL OTHER RES
(defun scr2log (dots &optional (target-res (cs-target-res)))
  (round (* dots (cs-logical-dpi))
         target-res))

(defun log2scr (logv &optional (target-res (cs-target-res)))
  (floor-round (* logv target-res )
         (cs-logical-dpi)))

(defun cs-archos-dpi ()
  (cs-logical-dpi))

(defun floor-round (x &optional (divisor 1))
  (ceiling (- (/ x divisor) 1/2)))

;converts logical pixels to screen pixels given the current target resolution OR OPTIONAL OTHER RES
(defun logical-to-screen-vector (dots &optional target-res)
  (let ((convert-res (or target-res (cs-target-res))))  
    (floor-round (* dots convert-res) (cs-logical-dpi))))

(defun logical-to-screen-point (point &optional target-res)
  (mkv2
   (log2scr (v2-h point) target-res)
   (log2scr (v2-v point) target-res)))

(defun screen-to-logical-v2 (point &optional target-res)
  (mkv2
   (scr2log (v2-h point) target-res)
   (scr2log (v2-v point) target-res)))

(defun nr-screen-to-logical (logical-rect screenrect &optional target-res)
  (nr-make logical-rect
   (scr2log (r-left screenrect) target-res)
   (scr2log (r-top screenrect) target-res)
   (scr2log (r-right screenrect) target-res)
    (scr2log (r-bottom screenrect) target-res)))

; logical-to-target is a more sensible name throughout

(defun logical-to-target-vector (dots &optional target-res)
  (log2scr dots target-res))
;--------------------------------------------------------------------------------------------

(defun r-logical-to-screen (logical-rect &optional target-res)
  (count-it :r-logical-to-screen)
  (nr-logical-to-screen (mkR 0 0 0 0) logical-rect target-res))

(defun nr-logical-to-screen (screenrect logical-rect &optional target-res)
  (nr-make screenrect
   (log2scr (r-left logical-rect) target-res)
   (log2scr (r-top logical-rect) target-res)
   (log2scr (r-right logical-rect) target-res)
    (log2scr (r-bottom logical-rect) target-res)))

;------------------------------------------------------------------------------------------------

;;;(defun set-scaling (window)
;;;  #+shh(trc "targetResolution" (targetRes window))
;;; 
;;;  (set-current-target-resolution (cs-logical-screen-resolution))          ;here and below, we'll probably make scalable
;;;  ;(set-current-target-resolution (cs-logical-dpi))
;;;  (let ((dc (device-context window))
;;;        (display-dpi (cs-logical-screen-resolution))                       ;... and use (targetRes window)
;;;        (logical-dpi (cs-logical-dpi)))
;;;     (os-SetMapMode dc win:MM_ISOTROPIC)
;;;     (os-SetWindowExtEx dc logical-dpi logical-dpi ct:hnull)                  
;;;     (os-SetViewportExtEx dc display-dpi display-dpi ct:hnull)))


(defun move-v2-x-y (v2 x y)
  (incf (v2-h v2) x)
  (incf (v2-v v2) y)
  v2)

(defmethod ncanvas-to-screen-point (self point)
  (ncanvas-to-screen-point (fm-parent self)
                          (move-v2-x-y point (px self) (py self))))

(defmethod res-to-res ((amount number) fromRes toRes)
  (if tores
      (round (* amount fromRes) toRes)
    fromres))

(defmethod res-to-res ((point v2) fromRes toRes)
  (nres-to-res (copy-v2 point) fromRes toRes))

#+no-2e-h
(defmethod nres-to-res ((point v2) fromRes tores)
  (setf (v2-h point) (res-to-res (v2-h point) fromRes toRes))
  (setf (v2-v point) (res-to-res (v2-v point) fromRes toRes))
  point)

(defmethod res-to-res ((box rect) fromRes toRes)
  (count-it :res-to-res)
  (nres-to-res (nr-copy (mkR 0 0 0 0) box) fromres toRes))

(defmethod nres-to-res :around (geothing fromRes (tores null))
  (declare (ignore fromres))
  geothing)

(defmethod nres-to-res ((box rect) fromRes tores)
  (setf (r-left box) (res-to-res (r-left box) fromRes toRes))
  (setf (r-top box) (res-to-res (r-top box) fromRes toRes))
  (setf (r-right box) (res-to-res (r-right box) fromRes toRes))
  (setf (r-bottom box) (res-to-res (r-bottom box) fromRes toRes))
  box)

(defun canvas-to-screen-box (self box)
  (count-it :canvas-to-screen-box)
  (nr-make-from-corners 
   (mkR 0 0 0 0)
   (ncanvas-to-screen-point self (r-top-left box))
   (ncanvas-to-screen-point self (r-bottom-right box))))

