;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cello; -*-
;;;
;;; Copyright  2004 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :cello)

; ------------- Frame classes and mk-macros

(defclass FrameStyle ()
    ((edges :initarg :edges :accessor edges)))

(defclass FrameRGB (FrameStyle)
    ((color :initarg :color :reader color :initform :black)
     (thickness :initarg :thickness :reader thickness :initform (uPts 0.75)) ; ML 000202 no longer hairline
     (penStyle :initarg :penStyle :reader penStyle :initform :PS_SOLID)))

(defclass Frame3D (FrameStyle)
  ((3dStyle :initarg :3dStyle :reader 3dStyle :initform :EDGE-RAISED)
   (thickness :initarg :thickness :initform (u8ths 1) :reader thickness)
   (of-window :initarg :of-window :initform nil :reader of-window)
   (w-box :initarg :w-box :initform nil :reader w-box)))

(defmacro mkFrame (&optional (edges :BF-RECT))
   `(make-instance 'Frame3D :edges (logior :BF-MONO ,edges) :3DStyle :EDGE-SUNKEN))

(defmacro mkFrame3D (3dStyle &optional (edges :BF-RECT))
   `(make-instance 'Frame3D :edges ,edges :3DStyle ,3DStyle))

(let ((normalv (make-ff-array :float 0 0 0)))
  (declare (ignorable normalv))
  (defun cgl-normal (x1 y1 z1 x2 y2 z2 x3 y3 z3)
    (declare (ignorable ncalc-normalf
                    x1 y1 z1
                    x2 y2 z2
                    x3 y3 z3))
    (gl-normal3fv (ncalc-normalf normalv
                    x1 y1 z1
                    x2 y2 z2
                    x3 y3 z3))))

(defun r-inset (r in)
  (mkr (+ (r-left r) in) 
    (+ (r-top r) (downs in))
    (- (r-right r) in)
    (+ (r-bottom r) (ups in))))
(defmodel skin ()
  ((texture-name :accessor texture-name :initarg :texture-name :initform nil)))

(defmethod ix-render-layer ((self Frame3d) lbox)
  (let* (
         (br (r-right lbox))
         (bl (r-left lbox))
         (bt (r-top lbox))
         (bb (r-bottom lbox))
         (thick (thickness self))
         (in thick)
         (skin (bwhen (w (of-window self))
                 (skin w)))
         (face (r-inset lbox in))
         uback uface
         (bz (ecase (3dstyle self)
               (:edge-sunken (outs thick))
               (:edge-raised (ins thick))))
         )
    ;; (print `(f3d-lbox=> ,lbox thick> ,thick))

    (if skin
      (let ((wr (gbox (of-window self)))
            (r (w-box self)))
        (gl-bind-texture GL_TEXTURE_2d (texture-name skin))
        (gl-Enable GL_TEXTURE_2d)
        (gl-Enable GL_MODULATE)
        (setf uback (r-unitize r wr))
        (setf uface (r-unitize (r-inset r in) wr))
        )
      (gl-disable GL_TEXTURE_2d))

    (flet
          ((vrtb ()
             (when skin
               (gl-tex-coord2f (r-right uback)(r-top uback)))
             (gl-vertex3f br bt bz))
           (vltb ()
             (when skin
               (gl-tex-coord2f (r-left uback)(r-top uback)))
             (gl-vertex3f bl bt bz))
           (vlbb ()
             (when skin
               (gl-tex-coord2f (r-left uback)(r-bottom uback)))
             (gl-vertex3f bl bb bz))
           (vrbb ()
             (when skin
               (gl-tex-coord2f (r-right uback)(r-bottom uback)))
             (gl-vertex3f br bb bz))
           (vltf ()
             (when skin
               (gl-tex-coord2f (r-left uface)(r-top uface)))
             (gl-vertex3f (r-left face) (r-top face) 0))
           (vlbf ()
             (when skin
               (gl-tex-coord2f (r-left uface)(r-bottom uface)))
             (gl-vertex3f (+ bl in) (+ bb in) 0))
           (vrtf ()
             (when skin
               (gl-tex-coord2f (r-right uface)(r-top uface)))
             (gl-vertex3f (- br in) (- bt in) 0))
           (vrbf ()
             (when skin
               (gl-tex-coord2f (r-right uface)(r-bottom uface)))                   
             (gl-vertex3f (- br in) (+ bb in) 0))
           )
        
        (within-gl-begin-end (GL_QUADS)
          ;; top
          (cgl-normal bl bt bz
                      br bt bz
                      (- br in) (+ bt (downs in)) 0)
          (vltb)(vrtb)(vrtf)(vltf)
          
          ;; left
          (cgl-normal (+ bl in) (+ bb (ups in)) 0
            bl bb bz
            bl bt bz)
          (vlbf)(vlbb)(vltb)(vltf)

          ;; bottom
          (cgl-normal 
           (- br in) (+ bb (ups in)) 0
           br bb bz
           bl bb bz)
          (vlbb)(vlbf)(vrbf)(vrbb)

          ;; right
          (cgl-normal (- br in) (+ bt (downs in)) 0
                      br bt bz
            br bb bz)
          (vrbb)(vrbf)(vrtf)(vrtb)

          ;; front
          (cgl-normal
           (- br in) (+ bt (downs in)) 0
           (+ bl in) (- bt in) 0
           (+ bl in) (+ bb in) 0
           )

          #+not (cgl-normal
           (+ bl in) (+ bb in) 0
           (+ bl in) (- bt in) 0
           (- br in) (+ bt (downs in)) 0)

          (vltf)(vrtf)(vrbf)(vlbf)

          )
        (ogb::glec :f3d)
        )))

(defmacro mkFrameRGB ((color &optional (edges :all)) &rest kwargs)
   ; probably need penstyle before long
   `(make-instance 'FrameRGB :color ,color :edges ,edges ,@kwargs))

(defclass Cone3D (Frame3d)())

(defmethod ix-render-layer ((self Cone3d) lbox)
  ;
  ; we expect to have been gl-translated to the center of
  ; the lbox parameter
  ;
  (let* ((r (floor (r-width lbox) 2)))

    (gl-translatef 0 0 -100)
    (gl-rotatef 45 0 0 0)
    (gl-scalef 1.1 1.1 1.1)
    (glut-solid-cone r (* 4 r)
      9 1)
    
    (ogb::glec :f3d)))

(defclass Sphere3D (Frame3d)())

(defmethod ix-render-layer ((self Sphere3d) lbox)
  (let* ((r (floor (r-width lbox) 2)))
    (trc nil "drawing sphere radius" self lbox r)
    (gl-translatef 0 0 1000)
    (gl-scalef 1.1 1.1 1.1)
    (glut-solid-sphere (* 100 r) 9 1)
    (ogb::glec :f3d)))