;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cello; -*-
;;;
;;; Copyright (c) 1995,2003 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.


(in-package :cello)

#|

Let's do something sick. An inspector needs something to inspect. Let's have it inspect
itself. Here is the "hello, world" of inspectors. As will be the case throughout this
document, the code is shown first, then its explnation.
|#

#+when-ready
(progn
  (defmodel inspector-window (Window)
    ((target :initarg :target :accessor target))
    (:default-initargs
        :title$ "Cello Inspector"
      :target (cv nil) ;; the object to be inspected goes here
      :layers (list +white+ :fill-lr)
      :kids (list
             (aLabel (c? (format nil "~a" (target .inspector.)))
               :pX 0 :pY 0))))
  
  (defmethod initialize-instance :after ((self inspector-window) &key)
    (unless (target self)
      (setf (target self) self))) ;; inspecting itself!

  (defun test-inspector ()
    (run-window (make-instance 'inspector-window))))

#|

Let's walk through the novel bits of code in order.

Defmodel: defclass plus the automatic generation of plumbing in support
of Cells. 

Window: the base Cello window class.

(cv nil): We will want to navigate amongst runtime data structures with the
same inspector window, so target will have to change. See C/B to
learn why such a creature must be encapsulated in a CV (c-variable). 

layers: part of how one specifies the appearance of any Cello Image instance.

C?: In Cell Basics we learned that this value gets computed just after a new model instance
gets added to the larger model tree, then re-computed as dependencies change.

aLabel: just a simple macro expanding into make-instance of a non-editable text widget

.w.: a symbol macro which expands into code to search up the runtime Family tree to the
first Window instance (in this case not far, to the label's immediate parent.

:pX,:pY: the Cello coordinate system. Like layers, not a concern in this document

Now we can go ahead and try out our
minimalist inspector...

(test-inspector)


You should see something like this.

<insert GRAPHIC>

the escape key or close icon will close the window.

Before enhancing the inspector, let's take a moment to add a convenient way to
inspect a GUI element of an interface on which one might be working, perhaps
to debug some problem with it.

It so happens the Cello API offers a do-right-button generic function callback
one can specialize on the widget clicked. We'll have control right-click inspect
any widget.

|#

#+when-ready
(defmethod do-right-button :around ((self Image) buttons wxwy)
  (declare (ignorable buttons wxwy))
  (if (not (control-key-down buttons))
      (call-next-method)
    (push (make-instance 'inspector-window
            :target (cv self))
      (kids *sys*))))

#|
New stuff:

control-key-down: Provided by the nascent Cello API.

*sys*: run-window (called above) initializes Cello, including stashing in *sys* 
an instance of the System class to serve as a root for the runtime 
model we will build. System is a subclass of Family (see C/B), whose kids slot forms a
recursive tree of runtime model instances. Ordinarily, the kids of a system
instance are windows.

(push [make window] (kids *sys*)): in C/B we learned that an echo function on the slot
kids causes new model instances to be activated Cell-wise. This in turn leads to the
new inspector window being initialized and displayed.

DO> run the above test function, then control right-click anywhere on the text widget
describing the inspector-window.

You should see something like this. <insert GRAPHIC>
 
Now let's show a list of slots. At the same time, let's break out
the rule for the inspector window content just to tidy up the code.
|#


#+when-ready
(progn
  (defmodel inspector-window (Window)
    ((target :initarg :target :accessor target))
    (:default-initargs
        :title$ "Cello Inspector"
      :target (cv nil)
      :layers (list +white+ :fill-lr)
      :content (inspector-window-content)))

  (defun inspector-window-content ()
    (c? (let ((tgt (^target)))
          (aStack (:pX 0 :pY 0)
            (aLabel (c? (format nil "~a" tgt)))
            (mapcar (lambda (slot)
                      (aLabel (format nil "~a"
                                (string-downcase
                                 (slot-definition-name slot)))))
              (class-slots (class-of tgt))))))))

#|
Whoa. That's ugly. And we need scroll bars, because we cannot see all the slots. 
And sorting on name would be nice. And so on. But functionality is more fun, so 
let's defer cosmetics for a moment.

First, let's see the slot-value as well as the slot name, and since we can
expect these slot displays to get interesting, let's break it out now in its 
own class.
|#

#+when-ready
(progn
  (defmodel slot-inspector (IXRow)
    ((slot :initarg :slot :reader slot))
    (:documentation "Inspector sub-view of a CLOS slot")
    (:default-initargs
        :spacing (u8ths 2)
      :kids (c? (let ((slot-name (slot-definition-name (^slot))))
                  (theKids
                   (aLabel (format nil "~a"
                             (string-downcase
                              slot-name)))
                   (aLabel (format nil "~a"
                             (slot-value (target .inspector.)
                               slot-name))))))))
    
    ;; and now:
    
  (defun inspector-window-content ()
    (c? (let ((tgt (^target)))
          (aStack (:pX 0 :pY 0 :spacing (u8ths 1))
            (aLabel (c? (format nil "~a" tgt)))
            (aStack (:pX (u8ths 1) :lR (C? (^fill-right Window)))
              (mapcar (lambda (slot)
                        (mkPart :slotview (slot-inspector)
                          :slot slot))
                (class-slots (class-of tgt))))))))
  )
  
#|
Oops, a real-world face-plant: Break: whoa. more than 32 used? i see 33
Remind me to fix that message, I never thought it would come up. It means some
cell is dependent on more than 32 other cells, a limitation in Cells arising from
pure laziness on my part (and the belief that I would get away with it). But...

...hey, a window has 78 slots at this writing, and the IXStack container happens to size itself
horizontally to fit the largest child. This was not a problem before because there is no
dependency in: (aLabel (format nil "~a" (string-downcase slot-name)))

...so the horizontal dimensions of the slot-name label were constant and cells do not
depend on constant other cells. But now we are showing the slot value, and that does have a 
dependency: (aLabel (format nil "~a" (slot-value (target .inspector.) slot-name)))

...so the stack dutifully records its dependency on 78 slots. I should just fix the internals,
but there is actually a mistake here which also contributes to the error. The stack need not 
shrink to fit the slot-inspectors, it can be as wide as the window, and that will be much more
efficient. So let's just do that and fix Cell internals when we really do depend on 33 other cells.

|#
#+nahh
(defun inspector-window-content ()
    (c? (let ((tgt (^target)))
          (aStack (:pX 0 :pY 0 :spacing (u8ths 1))
            (aLabel (c? (format nil "~a" tgt)))
            (aStack (:pX (u8ths 1) :lR (C? (^fill-right Window))) ; <= ignore kid widths, size to window
              (mapcar (lambda (slot)
                        (mkPart :slotview (slot-inspector)
                          :slot slot))
                (class-slots (class-of tgt))))))))

#|
OK, now let's allow the user to navigate from one
CLOS instance to another by left-clicking a slot. We'll need to
make our slot-inspector a Control to get clicks under the Cello API.

Warning: this will fail, and then we will work out how to fix it. I
left this in because the failure was Cell-based, and a new one on me.
I hope this provides a realistic look at life with Cells.

|#
  
#+this-produces-an-error
(defmodel slot-inspector (Control IXRow)
    ((slot :initarg :slot :reader slot))
    (:documentation "Inspector sub-view of a CLOS slot")
    (:default-initargs
        :spacing (u8ths 2)
      :controlaction (lambda (self event)
                       (declare (ignorable event))
                       (setf (target .inspector.)
                         (md-value (second (kids self)))))
      :kids (c? (let ((slot-name (slot-definition-name (^slot))))
                  (theKids
                   (aLabel (format nil "~a" (string-downcase slot-name)))
                   (aLabel (c? (format nil "~a" (^md-value)))
                     :md-value (c? (slot-value (target .inspector.) slot-name))))))))


#|
Whoa!!! Trouble: 

Error: The slot TICKFINE is missing from the object MGSYS of class
       #<STANDARD-CLASS MGSYSTEM> during operation SLOT-VALUE

Clearly a slot-inspector widget dedicated to the TICKFINE slot of the Window
instance we were just inspecting is still trying to access that slot of
the new instance, which is an MGSystem instance. How did this happen?

The overall design is OK. When the inspector looks at a new instance, a new
set of slot-inspectors will be generated for that instance. But each slot-inspector
references (md-value .w.), and so becomes dependent on that (unnecessarily, since
the slot-inspector dies when that value changes). So what went wrong?

Cell dataflow first changes a slot such as the inspector's md-value slot, then
notifies any dependents one by one. It turns out that the slot-inspectors got
told to refresh themselves before the GUI structure got told to make a whole
new set of slot-inspectors. So it looked for its assigned slot (TICKFINE) in
an instance not having such a slot.

We /really/ do not want to have to worry about the order in which Cell internals
work, so let's see if we can make this work. It turns out there are several ways to
avoid this. One would be:

(defun inspector-window-content ()
    (c? (let ((tgt (^md-value)))
          (aStack (:pX 0 :pY 0 :spacing (u8ths 1))
            (aLabel (c? (format nil "~a" tgt)))
            (aStack (:pX (u8ths 1))
              (mapcar (lambda (slot)
                        (mkPart :slotview (slot-inspector)
                          :si-value (slot-value tgt (slot-definition-name slot))
                          :md-value slot))
                (class-slots (class-of tgt))))))))

ie, pick the slot value out and specify it as an initarg to the slot-inspector. This
would also be faster, but then we rule out a future enhancement to have the inspector
automatically show any change to the inspected object. Another (untested!) fix would be:

(defun inspector-window-content ()
    (c? (aStack (:pX 0 :pY 0 :spacing (u8ths 1))
          (aLabel (c? (format nil "~a" (md-value .inspector.))))
          (aStack (:pX (u8ths 1))
            (let (kids)
              (dotimes (n 100 (nreverse kids))
                (push (mkPart :slotview (slot-inspector)
                        :md-value (c? (elt (class-slots (class-of (md-value .inspector.))) n)))
                  kids)))))))

That makes a fixed group of slot-inspectors which change on the fly which slot
they represent. This would happen to be a big speed improvement as one navigated
from instance to instance. Why would this not have the same problem, tho, as the
original formulation? Consider the rule which fell over last time:

        :si-value (c? (slot-value
                       (md-value .inspector.)
                       (slot-definition-name (^md-value))))

Cell internals will in fact address this rule first, but they will notice that
the dependency (^md-value) which is the slot-inspector's assigned slot is
affected by the same change, and is smart enough to re-evaluate that first. It is
a little complicated, but originally the slot did not see that it depended for
its very existence on the same change (a possible enhancement to Cells, now that
I have run into this class of error). So what to do?

Let's cheat. We realized right away that the automatic dependency arising from
the form (md-value .inspector.) was one we did not really need, so let's just defeat the
auto-dependency mechanism and move on. Later we will go for the more efficient
approach and avoid the issue that way.

|#

#+when-ready
(defmodel slot-inspector (Control IXRow)
      ((si-value :reader si-value :initarg :si-value))
      (:documentation "Inspector sub-view of a CLOS slot (the md-value)")
      (:default-initargs
          :spacing (u8ths 2)
        :si-value (c? (slot-value
                       (without-c-dependency (md-value .inspector.))
                       (slot-definition-name (^md-value))))
        :controlaction (lambda (self event)
                         (declare (ignorable event))
                         (setf (md-value .inspector.)
                           (si-value self)))
        :kids (slot-inspector-kids)))

#|
Uh-oh:
Error: windows from prior test remain extant. if not visible, evaluate:
       (setf (kids *sys*) nil)

This is a problem in iterative development I intend to eliminate soon 
because it is a tremendous nuisance, but for now just:

(setf (kids *sys*) nil)

...and run again.

Try clicking first on the .parent slot to inspect the System instance,
then from there click on the Mouse slot to see that Cello-implementing instance.
Cool, but now we need a way to back up, and while we are at it we may as well
add a way to go forward. So we need widgets for those commands and a slot
in which to keep a history. 

(continued in object-inspector-02)

|#