;; -*- mode: Lisp; Syntax: Common-Lisp; Package: cells; -*-
;;;
;;; Copyright (c) 1995,2003 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :cells)


(defun c-link-ex (used &aux (user (car *c-calculators*)))
  (c-assert user)
  
  (cond
   ((cmdead user) (return-from c-link-ex nil))
   ((null used)
    ;
    ; no cell on used value so it is constant, but if a synapse is involved the constant
    ; must still be filtered thru that, albeit only this once
    ;
    (when *synapse-factory*
      (c-assert (car *c-calculators*)) ;; sanity-check
      (funcall *synapse-factory* nil (car *c-calculators*))))

   ((or (not (typep used 'cell))
      (and (typep used 'c-dependent)
        (c-optimized-away-p used)))
    (return-from c-link-ex nil))

   (t 
    ;
    ; --------- debug stuff --------------
    (c-assert user)
    (c-assert (c-model user))
    (c-assert (c-model used))
    (c-assert (not (cmdead user)) () "dead user in link-ex ~a, used being ~a" user used)
    (c-assert (not (cmdead used)) () "dead used in link-ex ~a, user being ~a" used user)

    #+dfdbg (trc user "c-link > user, used" user used)
    (c-assert (not (eq :eternal-rest (md-state (c-model user)))))
    (c-assert (not (eq :eternal-rest (md-state (c-model used)))))
    (count-it :c-link-entry)
    (when *c-debug*
      (c-assert (or (null *synapse-factory*)
                (functionp *synapse-factory*))
        ()
        "~s is not a function, but was supplied as a synapse factory between ~s and ~s. probably parentheses wrong, as in (- (^lr x 96))"
        *synapse-factory* used user))
     
    (let ((used-link
           (or 
            ;;  check if linked already
            ;;  /// looks like a bug: cannot have two synaptic dependencies on same
            ;;  /// cell slot...probably need to "name" the synapses just for this purpose
            ;;
            (c-find-used-link user used)
            ;;
            ;;  following may have been a goof, but i like it: let synapse factory
            ;;  decide not to produce a synapse, in which case dumb direct c-cell link
            ;;  gets created.
            ;;
            (bwhen (syn (and *synapse-factory*
                          (funcall *synapse-factory* used user)))
              (c-add-user used syn)
              (c-add-used user syn)
              ;;(trc used "c-link>     users now:" (mapcar #'celltrueuser (c-users used)))
              (trc nil "setting used to syn" syn used)
              syn)
            ;;
            ;;  make dumb link: used just tells user to rethink.
            ;;
            (progn
              (trc nil "c-link > new user,used " user used)
              (c-add-user used user)
              (c-add-used user used)
              used))))
      
      (c-assert used-link)
      (c-assert (position used-link (cd-useds user))
        ()
        "used-link ~a does not appear in useds ~a of user ~a"
        used-link (cd-useds user) user)
      
      (let ((mapn (- *cd-usagect*
                    (- (length (cd-useds user))
                      (or (position used-link (cd-useds user)) 0)))))      
        ;; (trc user "c-link> setting usage bit" user mapn used-link)
        (if (minusp mapn)
            (break "whoa. more than ~d used? i see ~d" *cd-usagect* (length (cd-useds user)))
          (cd-usage-set user mapn)))
      used-link))))
   
(defun cd-usage-set (c mapn)
  (when (typep c 'synapse)
    (setf (syn-relevant c) t))
  (setf (sbit (cd-usage c) mapn) 1))

(defun cd-usage-clear-all (c)
  (bit-and (cd-usage c)
           #*0000000000000000000000000000000000000000000000000000000000000000
           t))

(defun c-find-used-link (user-cell used)
  "find any existing link to user-cell, the cell itself if direct or a synapse leading to it"
  (some (lambda (user)
          (if (typep user 'synapse)
              (when (eql user-cell (syn-user user))
                user) ;; the synapse is the used link
            (when (eql user-cell user)
              used))) ;; the link to used is direct (non-synaptic)
        (c-users used)))

(defun c-add-user (used user)
  (count-it :c-adduser)
  
  (typecase used
    (cell
     (trc nil "c-add-user conventional >  user, used" user used)
     (pushnew user (c-users used)))
    
    (synapse (setf (syn-user used) user)))
  
  used)

(defun c-user-path-exists-p (from-used to-user)
  (typecase from-used
    (synapse (c-user-path-exists-p (syn-user from-used) to-user))
    (cell
     (or (find to-user (c-users from-used))
       (find-if (lambda (from-used-user)
                  (c-user-path-exists-p (c-user-true from-used-user) to-user))
         (c-users from-used))))))

; -----------

(defun c-add-used (user used)
  (count-it :c-used)
  #+ucount (unless (member used (cd-useds user))
             (incf *cd-useds*)
             (when (zerop (mod *cd-useds* 100))
               (trc "useds count = " *cd-useds*)))
  (pushnew used (cd-useds user))
  (trc nil "c-add-used>  user <= used" user used (length (cd-useds user)))
  (mapcar 'c-users-resort (cd-useds user))
  (cd-useds user))

(defun c-users-resort (used)
  (typecase used
    (synapse (c-users-resort (syn-used used)))
    (cell
     (when (second (c-users used))
       (setf (c-users used) (sort (c-users used) 'c-user-path-exists-p))
       (trc nil "c-users-resort resorted users >  used" used (mapcar 'c-slot-name (c-users used)))
       (mapcar 'c-users-resort (c-useds used))))))

(defmethod c-useds (other) (declare (ignore other)))
(defmethod c-useds ((c c-dependent)) (cd-useds c))


(defun c-quiesce (c)
  (typecase c
    (cell 
     (trc nil "c-quiesce unlinking" c)
     (c-unlink-from-used c)
     (when (typep c 'cell)
       (dolist (user (c-users c))
         (c-unlink-user c user)))
     (c-pending-set c nil :c-quiesce)
      ;;;   (setf (c-waking-state c) nil)
      ;;;   (when (eql :rpthead (c-model c))
      (trc nil "cell quiesce nulled cell awake" c))))

;-------------------------
                     
(defmethod c-unlink-from-used ((user c-dependent))
  (dolist (used (cd-useds user))
    #+dfdbg (trc user "unlinking from used" user used)
    (c-unlink-user used user))
  ;; shouldn't be necessary (setf (cd-useds user) nil)
  )

(defmethod c-unlink-from-used (other)
  (declare (ignore other)))

;----------------------------------------------------------

(defmethod c-unlink-user ((used cell) user)
  #+dfdbg (trc user "user unlinking from used" user used)
  (setf (c-users used) (delete user (c-users used)))
  (c-unlink-used user used))

(defmethod c-unlink-user ((syn synapse)  user)
  (c-assert (eq user (syn-user syn)))
  (c-unlink-user (syn-used syn) syn)
  (setf (syn-user syn) nil) ;; gc-paranoia?
  )

;-----------------------------------------------------------


(defmethod c-unlink-used ((user c-dependent) used)
  (setf (cd-useds user) (delete used (cd-useds user))))

(defmethod c-unlink-used ((syn synapse) used)
  (c-assert (eq used (syn-used syn)))
  (setf (syn-used syn) nil)
  (c-unlink-used (syn-user syn) syn))

; --- very low-vel abstraction

(defmethod c-user-true (c) c)
(defmethod c-user-true ((syn synapse)) (syn-user syn))
(defmethod c-used-true (c) c)
(defmethod c-used-true ((syn synapse)) (syn-used syn))
