;; -*- mode: Lisp; Syntax: Common-Lisp; Package: ogb; -*-
;;;
;;; Copyright  1995,2003 by Kenneth William Tilton.
;;;
;;; Permission is hereby granted, free of charge, to any person obtaining a copy 
;;; of this software and associated documentation files (the "Software"), to deal 
;;; in the Software without restriction, including without limitation the rights 
;;; to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
;;; copies of the Software, and to permit persons to whom the Software is furnished 
;;; to do so, subject to the following conditions:
;;;
;;; The above copyright notice and this permission notice shall be included in 
;;; all copies or substantial portions of the Software.
;;;
;;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
;;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
;;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
;;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
;;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
;;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS 
;;; IN THE SOFTWARE.

(in-package :ogb)

(defconstant WCX 640)        ;; Window Width
(defconstant WCY -480)        ;; Window Height
(defparameter g_rot 0.0f0)

;; Our GL Specific Initializations
(defun init()
  (gl-clear-depth 1.0)

  (gl-Shade-Model GL_SMOOTH)					;; Enable Smooth Shading
  (glClearColor 1.0  0.0  0.0  0.5)		;; Black Background
  (gl-Enable GL_DEPTH_TEST)					;; Enables Depth Testing
  (gl-Depth-Func GL_LEQUAL)					;; The Type Of Depth Testing To Do
  (gl-Hint GL_PERSPECTIVE_CORRECTION_HINT GL_NICEST)
  (gl-Enable GL_LIGHT0)						;; Enable Default Light (Quick And Dirty)
  (gl-Enable GL_LIGHTING)					;; Enable Lighting
  (gl-Enable GL_COLOR_MATERIAL)				;; Enable Coloring Of Material
  t)

(defvar *opengl-window-lock* #+lispworks (mp:make-lock :name "OpenGL"))
(defparameter *render-lock-window* nil)

(defmacro with-render-lock ((glutW) &body body)
  (let ((fn (gensym)))
    `(let ((,fn (lambda () ,@body)))
       (call-with-render-lock ,glutW ,fn))))

(defun call-with-render-lock (glutW fn)
  (let (lock)
    (unwind-protect
        (let ((lock (render-lock-seek glutw)))
          (when lock
            (funcall fn)))
      (render-lock-release lock))))

(defmethod render-lock-seek ( glutW )
  (when (plusp glutW)
    #-lispworks
    t
    #+lispworks
    (if (eq mp:*current-process* (mp:lock-owner *opengl-window-lock*))
        (cond
         ((eq *render-lock-window* glutW) :nested) ;; cool

         ((null *render-lock-window*)
          (break "OpenGL error: Reentrant call to start-rendering in process: ~a" glutW)
          nil)
         
         (t
          (print (list "OpenGL error: Nested call to start-rendering on different window: current : ~s, requested ~s" 
            *render-lock-window* glutW))
          nil))
    
      (progn
        (mp:process-lock *opengl-window-lock*)
        (setf *render-lock-window* glutW)
        :lock))))

(defmethod render-lock-release (lock)
  (when (eq lock :lock)       
    #+lispworks
    (progn
      (setf *render-lock-window* nil)
      (mp:process-unlock *opengl-window-lock*))))

#+lispworks
(FLI:DEFINE-FOREIGN-FUNCTION (lw-clear-color "glClearColor" :SOURCE)
    ((RED GLCLAMPF) (GREEN GLCLAMPF) (BLUE GLCLAMPF) (ALPHA GLCLAMPF))
  :MODULE "OPENGL32"
  :RESULT-TYPE :VOID
  :LANGUAGE :ANSI-C
  :CALLING-CONVENTION :stdcall)

(ff-defun-callable :cdecl :void nh14disp ()
  (nh14-disp))


(defun nh14-disp ()

  (gl-Load-Identity)						;; Reset The Current Modelview Matrix
  (gl-clear-Color 0.0 0.0 0.0  0.5)
  (gl-Clear (+ GL_COLOR_BUFFER_BIT GL_DEPTH_BUFFER_BIT))
          
  (gl-Translatef 0.0f0 0.0f0 2.0f0)			;; Move Into The Screen

  (font-glut-preview)

  (gl-rotatef g_rot 1.0f0 0.0f0 0.0f0)			;; Rotate On The X Axis
  (gl-rotatef (* g_rot 1.5f0) 0.0f0 1.0f0 0.0f0)	;; Rotate On The Y Axis
  (gl-rotatef (* g_rot 1.4f0) 0.0f0 0.0f0 1.0f0)	;; Rotate On The Z Axis
  (gl-Scalef 0.002  0.003  0.0)

  ;; Pulsing Colors Based On The Rotation
  (gl-Color3f (* 1.0f0 (cos (/ g_rot 20.0f0)))
          (* 1.0f0 (sin (/ g_rot 25.0f0)))
          (- 1.0f0 (* 0.5f0 (cos (/ g_rot 17.0f0)))))

  (gl-push-matrix)
          
  (gl-line-width 3)
  (ogb::glut-stroke-string (make-ff-pointer GLUT_STROKE_ROMAN)
    (format nil "NeHe - ~a" (/ g_rot 50.0)))
  (gl-pop-matrix)
  
  (glut-Swap-Buffers)
        
  (incf g_rot 0.4f0)
  (glut-post-redisplay))


(defun font-glut-preview ()
   (gl-push-matrix)
   
   (gl-color3f 1 1 1)
   (gl-Scalef 0.007  0.007  0.0)
   (loop for bitmap-font in
            '(glut_bitmap_8_by_13 glut_bitmap_9_by_15 
                glut_bitmap_helvetica_10 glut_bitmap_helvetica_12 glut_bitmap_helvetica_18
                glut_bitmap_times_roman_10 glut_bitmap_times_roman_24)
         for id = (symbol-value bitmap-font)
         for y-pos = 0 then (round (- y-pos (glut-bitmap-height id)))
            do
            (assert (numberp id))
            ;(print (list bitmap-font (glut-bitmap-height id) y-pos id))
            (gl-raster-pos3i -250 y-pos 0)
            (glut-bitmap-string id
               (format nil "Hello, ~a" bitmap-font)))
   (gl-pop-matrix)
   (gl-push-matrix)
   (gl-translatef -2 1 0)
   (gl-Scalef 0.001  0.001  0.0)
   (gl-line-width 3)
   (loop for stroke-font in
                       '(glut_stroke_mono_roman glut_stroke_roman)
                    for id = (symbol-value stroke-font)
                    for y-pos = 0 then (round (- y-pos (/ (glut-stroke-height id) 1)))
                       do
                       (assert (numberp id))
                       ;(print (list stroke-font (glut-stroke-height id) y-pos id))
                       (gl-translatef 0 y-pos 0)
                       
                       (let ((msg (format nil "Hello, ~a  ~a"  (round (glut-stroke-height id))
                                           stroke-font)))
                         (with-cstring (cc msg)
                             (glut-stroke-string id msg)
                             (gl-translatef (- (glut-stroke-length (make-ff-pointer id) cc))
                                0 0))))
   (gl-pop-matrix))

#+test
(lesson-14)

(defun lesson-14 ()

  (let ((*glut-lib-path* "c:/dvx/lib/freeglut.dll"))
    (ogb-glut-init nil))
  (glut-set-option GLUT_ACTION_ON_WINDOW_CLOSE GLUT_ACTION_GLUTMAINLOOP_RETURNS)
  
  (glut-Init-Display-Mode (+ GLUT_RGB GLUT_DOUBLE)) ;; Display Mode (Rgb And Double Buffered)
  (glut-Init-Window-Size WCX WCY)   ;; Window Size If We Start In Windowed Mode

  (let ((key "NeHe's OpenGL Framework"))
    (uffi:with-cstring (key-native key)
      (glut-Create-Window key-native)))
  
  (init) ;                                          // Our Initialization
  ;; Set up the callbacks in OpenGL/GLUT
  (glut-Display-Func (ff-register-callable 'nh14disp))
  (glut-wm-close-Func (ff-register-callable 'mgwclose))
  (glut-Keyboard-Func (ff-register-callable 'mgwkey))

  (gl-matrix-mode GL_PROJECTION)
  (gl-Load-Identity)
  (glu-Perspective 70d0 1d0 1d0 1000d0)
  (glu-Look-At 0d0 0d0 5d0 0d0 0d0 0d0 0d0 1d0 0d0)

  (gl-Matrix-Mode GL_MODELVIEW)
  (gl-Load-Identity)


  (gl-Clear-Depth 1d0)

  #+this-works-to-but-pegs-cpu
  (glutMainLoop)
  (do ()
      ((zerop (glut-get-window)))
    ;;(format t "before main loop ~a | ~&" (glut-get-window))
    (progn ;; with-render-lock ((glut-get-window))
      (glutMainLoopEvent))
    (sleep 0.01)
    ))

#+test
(lesson-14)

(ff-defun-callable :cdecl :void mgwkey (k x y)
  (print 'bingo)
  (mgwkeyi k x y))

(defun mgwkeyi (k x y)
  (declare (ignorable k x y))
  (print (list "mgwkey" k x y (glutgetwindow)))
  (let ((mods (glut-get-modifiers)))
    (declare (ignorable mods))
    (print (list :keyboard k mods x  y (code-char (logand k #xff))#+not(glut-get-window)))
    (case (code-char (logand k #xff))
      (#\Escape (progn
                  (print (list "destroying window" (glutgetwindow) )
                    )
                  (glut-Destroy-Window (glutgetwindow)))))))